from typing import Any, Dict

import pytest
from packaging.version import Version
from scrapy import __version__ as SCRAPY_VERSION

if Version(SCRAPY_VERSION) < Version("2.7"):
    pytest.skip("Skipping tests for Scrapy ≥ 2.7", allow_module_level=True)

from scrapy import Request, Spider
from scrapy.utils.misc import create_instance

from scrapy_zyte_api import ScrapyZyteAPIRequestFingerprinter

from . import get_crawler

try:
    import scrapy_poet
except ImportError:
    scrapy_poet = None


def test_cache():
    crawler = get_crawler()
    fingerprinter = create_instance(
        ScrapyZyteAPIRequestFingerprinter, settings=crawler.settings, crawler=crawler
    )
    request = Request("https://example.com", meta={"zyte_api": True})
    fingerprint = fingerprinter.fingerprint(request)

    fingerprinter._param_parser = None  # Prevent later calls from working
    cached_fingerprint = fingerprinter.fingerprint(request)

    assert fingerprint == cached_fingerprint
    assert fingerprint == fingerprinter._cache[request]


def test_fallback_custom(caplog):
    class CustomFingerprinter:
        def fingerprint(self, request):
            return b"foo"

    settings = {
        "ZYTE_API_FALLBACK_REQUEST_FINGERPRINTER_CLASS": CustomFingerprinter,
    }
    crawler = get_crawler(settings)
    with caplog.at_level("WARNING"):
        fingerprinter = create_instance(
            ScrapyZyteAPIRequestFingerprinter,
            settings=crawler.settings,
            crawler=crawler,
        )
    request = Request("https://example.com")
    assert fingerprinter.fingerprint(request) == b"foo"
    request = Request("https://example.com", meta={"zyte_api": True})
    assert fingerprinter.fingerprint(request) != b"foo"
    try:
        import scrapy_poet  # noqa: F401
    except ImportError:
        pass
    else:
        assert (
            "does not point to a subclass of scrapy_poet.ScrapyPoetRequestFingerprinter"
            in caplog.text
        )


def test_fallback_default():
    crawler = get_crawler()
    fingerprinter = crawler.request_fingerprinter
    fallback_fingerprinter = (
        crawler.request_fingerprinter._fallback_request_fingerprinter
    )

    request = Request("https://example.com")
    new_fingerprint = fingerprinter.fingerprint(request)
    old_fingerprint = fallback_fingerprinter.fingerprint(request)
    assert new_fingerprint == old_fingerprint

    request = Request("https://example.com", meta={"zyte_api_automap": True})
    new_fingerprint = fingerprinter.fingerprint(request)
    assert old_fingerprint == fallback_fingerprinter.fingerprint(request)
    assert new_fingerprint != old_fingerprint


def test_headers():
    crawler = get_crawler()
    fingerprinter = create_instance(
        ScrapyZyteAPIRequestFingerprinter, settings=crawler.settings, crawler=crawler
    )
    request1 = Request(
        "https://example.com",
        meta={
            "zyte_api": {
                "customHttpRequestHeaders": [{"name": "foo", "value": "bar"}],
                "requestHeaders": {"referer": "baz"},
            }
        },
    )
    request2 = Request("https://example.com", meta={"zyte_api": True})
    fingerprint1 = fingerprinter.fingerprint(request1)
    fingerprint2 = fingerprinter.fingerprint(request2)
    assert fingerprint1 == fingerprint2


@pytest.mark.parametrize(
    "url,params,fingerprint",
    (
        (
            "https://example.com",
            {},
            b"\xccz|-\x1c%\xc5\xa3\x813\x91\x1a\x1a<\x95\x91\xf91a\n",
        ),
        (
            "https://example.com/a",
            {},
            b"x!<\xc5\x88\x08#\x9e\xf0\x19J\xd4\x92\x88\xb9\xb9\xce}\xb5\xda",
        ),
        (
            "https://example.com?a",
            {},
            b'\x80D\xdag"E\x8d=\xc7\xd68\xe1\xfd\xfd\x91\xe8\xd2.\xe6\xe4',
        ),
        (
            "https://example.com?a=b",
            {},
            b"r\xa6\x93\xa59\xb8\xb0\x9a\x90`p\xbf8\xdbW\x0f%\x17@N",
        ),
        (
            "https://example.com?a=b&a",
            {},
            b"T\x88[O\x8f\x87\xc1\xbb\x0e\xa3\xfbg^s\xf9=\x92?\x17\xe8",
        ),
        (
            "https://example.com?a=b&a=c",
            {},
            b"\xff=\xc3\xe74`\x048\xecM\xa3\xe8&\xb9\x06\xdf\xb2\xb0\x96\x8e",
        ),
        (
            "https://example.com",
            {"httpRequestBody": "Zm9v"},
            b";*\xa9Wt\xcfcso2\x9e\xa5\xd9_\xcc~_\xf5\\\xcd",
        ),
        (
            "https://example.com",
            {"httpRequestMethod": "POST"},
            b"\xe1\xf3&2R%\x0c\x82mf\x88E\x11L\x05w+\xa6V\xcb",
        ),
        (
            "https://example.com",
            {"httpResponseBody": True},
            b"e\x1e\xd3J0ya_\xca\xc3\xa0\xbe'h\x0ff*\xa6b\xf2",
        ),
        (
            "https://example.com",
            {"httpResponseHeaders": True},
            b"\xcc^\x0e$\xa7\xe5\x97\xb8\xbf\x7f0\xa3\xec\xf5B\\\xe1h\x1c\xee",
        ),
        (
            "https://example.com",
            {"browserHtml": True},
            b"\xb2\x8e\x98\xa9\xa2\xf2\xa6\x96\x01\xf6\x1dYa\xf7\xdf\xc2\xe5>x\x11",
        ),
        (
            "https://example.com",
            {"screenshot": True},
            b"\x8a\xd1\x1fut\x99\xf1\xc4\xcc\xa8\xfd\xd9\x7f\x1fY\xf8\xdf/'\xb3",
        ),
        (
            "https://example.com",
            {"screenshotOptions": {"format": "png"}},
            b"\xe2\xba\xeb\x16\xb9\xd4\x117\x19\xac\x7f\xb3\x17\xf5\xf6\xfc\x9e\x94l\xcf",
        ),
        (
            "https://example.com",
            {"geolocation": "US"},
            b"#\xe2\\\xce\xb88\xf8\xb4\x19\xa09KL\xe4\x87\x80\x00\x00A7",
        ),
        (
            "https://example.com",
            {"javascript": False},
            b"\x1c!\x89\xfc\xadd\xb3\xbf-_\x97\xca\xc0g\xbdo\xee\xdc\xdfo",
        ),
        (
            "https://example.com",
            {"actions": [{"action": "click", "selector": ".button"}]},
            b"\x83\xfa\x04\xfal\xc6d(\xe1\x06\xf1>b\xed\xbe\xb1\xf2\xac5E",
        ),
    ),
)
def test_known_fingerprints(url, params, fingerprint):
    """Test that known fingerprints remain the same, i.e. make sure that we do
    not accidentally modify fingerprints with future implementation changes."""
    crawler = get_crawler()
    fingerprinter = create_instance(
        ScrapyZyteAPIRequestFingerprinter, settings=crawler.settings, crawler=crawler
    )
    request = Request(url, meta={"zyte_api": params})
    actual_fingerprint = fingerprinter.fingerprint(request)
    assert actual_fingerprint == fingerprint


def test_metadata():
    settings = {"JOB": "1/2/3"}
    crawler = get_crawler(settings)
    job_fingerprinter = create_instance(
        ScrapyZyteAPIRequestFingerprinter, settings=crawler.settings, crawler=crawler
    )

    crawler = get_crawler()
    no_job_fingerprinter = create_instance(
        ScrapyZyteAPIRequestFingerprinter, settings=crawler.settings, crawler=crawler
    )

    request1 = Request("https://example.com", meta={"zyte_api": {"echoData": "foo"}})
    request2 = Request("https://example.com", meta={"zyte_api": True})

    fingerprint1 = job_fingerprinter.fingerprint(request1)
    fingerprint2 = job_fingerprinter.fingerprint(request2)
    fingerprint3 = no_job_fingerprinter.fingerprint(request1)
    fingerprint4 = no_job_fingerprinter.fingerprint(request2)

    assert fingerprint1 != fingerprint2
    assert fingerprint2 != fingerprint3
    assert fingerprint3 != fingerprint4
    assert fingerprint1 == fingerprint3
    assert fingerprint2 == fingerprint4


@pytest.mark.skipif(
    scrapy_poet is not None,
    reason=("scrapy-poet is installed, and test_deps already covers these scenarios"),
)
def test_only_end_parameters_matter():
    """Test that it does not matter how a request comes to use some Zyte API
    parameters, that the fingerprint is the same if the parameters actually
    sent to Zyte API are the same."""

    settings: Dict[str, Any] = {
        "ZYTE_API_TRANSPARENT_MODE": True,
    }
    crawler = get_crawler(settings)
    transparent_fingerprinter = crawler.request_fingerprinter

    crawler = get_crawler()
    default_fingerprinter = crawler.request_fingerprinter

    request = Request("https://example.com")
    fingerprint1 = transparent_fingerprinter.fingerprint(request)
    fingerprint2 = default_fingerprinter.fingerprint(request)

    raw_request = Request(
        "https://example.com",
        meta={
            "zyte_api": {
                "httpResponseBody": True,
                "httpResponseHeaders": True,
                "experimental": {
                    "responseCookies": True,
                },
            }
        },
    )
    fingerprint3 = transparent_fingerprinter.fingerprint(raw_request)
    fingerprint4 = default_fingerprinter.fingerprint(raw_request)

    auto_request = Request("https://example.com", meta={"zyte_api_automap": True})
    fingerprint5 = transparent_fingerprinter.fingerprint(auto_request)
    fingerprint6 = default_fingerprinter.fingerprint(auto_request)

    assert fingerprint1 != fingerprint2

    assert fingerprint3 == fingerprint4
    assert fingerprint5 == fingerprint6

    assert fingerprint1 == fingerprint3
    assert fingerprint1 == fingerprint5


@pytest.mark.parametrize(
    "url1,url2,match",
    (
        (
            "https://example.com",
            "https://example.com",
            True,
        ),
        (
            "https://example.com",
            "https://example.com/",
            True,
        ),
        (
            "https://example.com/a",
            "https://example.com/b",
            False,
        ),
        (
            "https://example.com/?1",
            "https://example.com/?2",
            False,
        ),
        (
            "https://example.com/?a=1&b=2",
            "https://example.com/?b=2&a=1",
            True,
        ),
        (
            "https://example.com",
            "https://example.com#",
            True,
        ),
        (
            "https://example.com#",
            "https://example.com#1",
            True,
        ),
        (
            "https://example.com#1",
            "https://example.com#2",
            True,
        ),
    ),
)
def test_url(url1, url2, match):
    crawler = get_crawler()
    fingerprinter = create_instance(
        ScrapyZyteAPIRequestFingerprinter, settings=crawler.settings, crawler=crawler
    )
    request1 = Request(url1, meta={"zyte_api_automap": True})
    fingerprint1 = fingerprinter.fingerprint(request1)
    request2 = Request(url2, meta={"zyte_api_automap": True})
    fingerprint2 = fingerprinter.fingerprint(request2)
    if match:
        assert fingerprint1 == fingerprint2
    else:
        assert fingerprint1 != fingerprint2


def merge_dicts(*dicts):
    return {k: v for d in dicts for k, v in d.items()}


@pytest.mark.parametrize(
    "params,match",
    (
        # As long as browserHtml or screenshot are True, different fragments
        # make for different fingerprints, regardless of other parameters. Same
        # for extraction types if browserHtml is set in *Options.extractFrom.
        *(
            (
                merge_dicts(body, headers, unknown, browser),
                False,
            )
            for body in (
                {},
                {"httpResponseBody": False},
                {"httpResponseBody": True},
            )
            for headers in (
                {},
                {"httpResponseHeaders": False},
                {"httpResponseHeaders": True},
            )
            for unknown in (
                {},
                {"unknown": False},
                {"unknown": True},
            )
            for browser in (
                {"browserHtml": True},
                {"screenshot": True},
                {"browserHtml": True, "screenshot": False},
                {"browserHtml": False, "screenshot": True},
                {"browserHtml": True, "screenshot": True},
                {"product": True, "productOptions": {"extractFrom": "browserHtml"}},
            )
        ),
        # If neither browserHtml nor screenshot are enabled, different
        # fragments do *not* make for different fingerprints. Same for
        # extraction types if browserHtml is not set in # *Options.extractFrom.
        *(
            (
                merge_dicts(body, headers, unknown, browser),
                True,
            )
            for body in (
                {},
                {"httpResponseBody": False},
                {"httpResponseBody": True},
            )
            for headers in (
                {},
                {"httpResponseHeaders": False},
                {"httpResponseHeaders": True},
            )
            for unknown in (
                {},
                {"unknown": False},
                {"unknown": True},
            )
            for browser in (
                {},
                {"browserHtml": False},
                {"screenshot": False},
                {"browserHtml": False, "screenshot": False},
                {"product": True},
                {
                    "product": True,
                    "productOptions": {"extractFrom": "httpResponseBody"},
                },
            )
        ),
    ),
)
def test_url_fragments(params, match):
    crawler = get_crawler()
    fingerprinter = create_instance(
        ScrapyZyteAPIRequestFingerprinter, settings=crawler.settings, crawler=crawler
    )
    request1 = Request("https://toscrape.com#1", meta={"zyte_api": params})
    fingerprint1 = fingerprinter.fingerprint(request1)
    request2 = Request("https://toscrape.com#2", meta={"zyte_api": params})
    fingerprint2 = fingerprinter.fingerprint(request2)
    if match:
        assert fingerprint1 == fingerprint2
    else:
        assert fingerprint1 != fingerprint2


def test_extract_types():
    crawler = get_crawler()
    fingerprinter = create_instance(
        ScrapyZyteAPIRequestFingerprinter, settings=crawler.settings, crawler=crawler
    )
    request1 = Request("https://toscrape.com", meta={"zyte_api": {"product": True}})
    fingerprint1 = fingerprinter.fingerprint(request1)
    request2 = Request(
        "https://toscrape.com", meta={"zyte_api": {"productNavigation": True}}
    )
    fingerprint2 = fingerprinter.fingerprint(request2)
    assert fingerprint1 != fingerprint2


def test_request_body():
    crawler = get_crawler()
    fingerprinter = create_instance(
        ScrapyZyteAPIRequestFingerprinter, settings=crawler.settings, crawler=crawler
    )
    request1 = Request(
        "https://toscrape.com", meta={"zyte_api": {"httpRequestBody": "Zm9v"}}
    )
    fingerprint1 = fingerprinter.fingerprint(request1)
    request2 = Request(
        "https://toscrape.com", meta={"zyte_api": {"httpRequestText": "foo"}}
    )
    fingerprint2 = fingerprinter.fingerprint(request2)
    assert fingerprint1 == fingerprint2


@pytest.mark.skipif(scrapy_poet is None, reason="scrapy-poet is not installed")
def test_deps():
    """Test that some injected dependencies do not affect fingerprinting at
    all (e.g. HttpClient) while others do (e.g. WebPage)."""
    from web_poet import HttpClient, WebPage

    request = Request("https://example.com")
    raw_request = Request(
        "https://example.com",
        meta={
            "zyte_api": {
                "httpResponseBody": True,
                "httpResponseHeaders": True,
                "experimental": {
                    "responseCookies": True,
                },
            }
        },
    )
    auto_request = Request("https://example.com", meta={"zyte_api_automap": True})

    class DepsSpider(Spider):
        name = "deps"

        def __init__(self, *args, **kwargs):
            self.client_request = Request(
                "https://example.com", callback=self.parse_client
            )
            self.client_raw_request = Request(
                "https://example.com",
                callback=self.parse_client,
                meta={
                    "zyte_api": {
                        "httpResponseBody": True,
                        "httpResponseHeaders": True,
                        "experimental": {
                            "responseCookies": True,
                        },
                    }
                },
            )
            self.client_auto_request = Request(
                "https://example.com",
                callback=self.parse_client,
                meta={"zyte_api_automap": True},
            )

            self.page_request = Request("https://example.com", callback=self.parse_page)
            self.page_raw_request = Request(
                "https://example.com",
                callback=self.parse_page,
                meta={
                    "zyte_api": {
                        "httpResponseBody": True,
                        "httpResponseHeaders": True,
                        "experimental": {
                            "responseCookies": True,
                        },
                    }
                },
            )
            self.page_auto_request = Request(
                "https://example.com",
                callback=self.parse_page,
                meta={"zyte_api_automap": True},
            )

        async def parse_client(self, response, a: HttpClient):
            pass

        async def parse_page(self, response, a: WebPage):
            pass

    default_crawler = get_crawler(spider_cls=DepsSpider)
    default_fingerprinter = default_crawler.request_fingerprinter
    transparent_crawler = get_crawler(
        {"ZYTE_API_TRANSPARENT_MODE": True}, spider_cls=DepsSpider
    )
    transparent_fingerprinter = transparent_crawler.request_fingerprinter

    request_default_fp = default_fingerprinter.fingerprint(request)
    request_transparent_fp = transparent_fingerprinter.fingerprint(request)
    raw_request_default_fp = default_fingerprinter.fingerprint(raw_request)
    raw_request_transparent_fp = transparent_fingerprinter.fingerprint(raw_request)
    auto_request_default_fp = default_fingerprinter.fingerprint(auto_request)
    auto_request_transparent_fp = transparent_fingerprinter.fingerprint(auto_request)

    client_request_default_fp = default_fingerprinter.fingerprint(
        default_crawler.spider.client_request
    )
    client_request_transparent_fp = transparent_fingerprinter.fingerprint(
        transparent_crawler.spider.client_request
    )
    client_raw_request_default_fp = default_fingerprinter.fingerprint(
        default_crawler.spider.client_raw_request
    )
    client_raw_request_transparent_fp = transparent_fingerprinter.fingerprint(
        transparent_crawler.spider.client_raw_request
    )
    client_auto_request_default_fp = default_fingerprinter.fingerprint(
        default_crawler.spider.client_auto_request
    )
    client_auto_request_transparent_fp = transparent_fingerprinter.fingerprint(
        transparent_crawler.spider.client_auto_request
    )

    page_request_default_fp = default_fingerprinter.fingerprint(
        default_crawler.spider.page_request
    )
    page_request_transparent_fp = transparent_fingerprinter.fingerprint(
        transparent_crawler.spider.page_request
    )
    page_raw_request_default_fp = default_fingerprinter.fingerprint(
        default_crawler.spider.page_raw_request
    )
    page_raw_request_transparent_fp = transparent_fingerprinter.fingerprint(
        transparent_crawler.spider.page_raw_request
    )
    page_auto_request_default_fp = default_fingerprinter.fingerprint(
        default_crawler.spider.page_auto_request
    )
    page_auto_request_transparent_fp = transparent_fingerprinter.fingerprint(
        transparent_crawler.spider.page_auto_request
    )

    assert request_default_fp != request_transparent_fp
    assert request_default_fp != raw_request_default_fp
    assert request_default_fp != raw_request_transparent_fp
    assert request_default_fp != auto_request_default_fp
    assert request_default_fp != auto_request_transparent_fp
    assert request_default_fp == client_request_default_fp
    assert request_default_fp != client_request_transparent_fp
    assert request_default_fp != client_raw_request_default_fp
    assert request_default_fp != client_raw_request_transparent_fp
    assert request_default_fp != client_auto_request_default_fp
    assert request_default_fp != client_auto_request_transparent_fp
    assert request_default_fp != page_request_default_fp
    assert request_default_fp != page_request_transparent_fp
    assert request_default_fp != page_raw_request_default_fp
    assert request_default_fp != page_raw_request_transparent_fp
    assert request_default_fp != page_auto_request_default_fp
    assert request_default_fp != page_auto_request_transparent_fp

    assert request_transparent_fp == raw_request_default_fp
    assert request_transparent_fp == raw_request_transparent_fp
    assert request_transparent_fp == auto_request_default_fp
    assert request_transparent_fp == auto_request_transparent_fp
    assert request_transparent_fp == client_request_transparent_fp
    assert request_transparent_fp == client_raw_request_default_fp
    assert request_transparent_fp == client_raw_request_transparent_fp
    assert request_transparent_fp == client_auto_request_default_fp
    assert request_transparent_fp == client_auto_request_transparent_fp
    assert request_transparent_fp != page_request_default_fp
    assert request_transparent_fp != page_request_transparent_fp
    assert request_transparent_fp != page_raw_request_default_fp
    assert request_transparent_fp != page_raw_request_transparent_fp
    assert request_transparent_fp != page_auto_request_default_fp
    assert request_transparent_fp != page_auto_request_transparent_fp

    assert page_request_default_fp != page_request_transparent_fp
    assert page_request_default_fp != page_raw_request_default_fp
    assert page_request_default_fp != page_raw_request_transparent_fp
    assert page_request_default_fp != page_auto_request_default_fp
    assert page_request_default_fp != page_auto_request_transparent_fp

    assert page_request_transparent_fp == page_raw_request_default_fp
    assert page_request_transparent_fp == page_raw_request_transparent_fp
    assert page_request_transparent_fp == page_auto_request_default_fp
    assert page_request_transparent_fp == page_auto_request_transparent_fp


def test_page_params():
    no_params_request = Request("https://example.com")
    empty_params_request = Request("https://example.com", meta={"page_params": {}})
    some_param_request = Request(
        "https://example.com", meta={"page_params": {"a": "b"}}
    )
    other_param_request = Request(
        "https://example.com", meta={"page_params": {"c": "d"}}
    )

    crawler = get_crawler({"ZYTE_API_TRANSPARENT_MODE": True})
    fingerprinter = crawler.request_fingerprinter

    no_params_fingerprint = fingerprinter.fingerprint(no_params_request)
    empty_params_fingerprint = fingerprinter.fingerprint(empty_params_request)
    some_param_fingerprint = fingerprinter.fingerprint(some_param_request)
    other_param_fingerprint = fingerprinter.fingerprint(other_param_request)

    if scrapy_poet is None:
        assert no_params_fingerprint == empty_params_fingerprint
        assert no_params_fingerprint == some_param_fingerprint
        assert no_params_fingerprint == other_param_fingerprint
    else:
        assert no_params_fingerprint == empty_params_fingerprint
        assert no_params_fingerprint != some_param_fingerprint
        assert no_params_fingerprint != other_param_fingerprint
        assert some_param_fingerprint != other_param_fingerprint



from base64 import b64decode

import attrs


class Geolocation:
    """A page input that forces a given geolocation for all other page inputs.

    The target geolocation must be :ref:`specified with an annotation
    <geolocation>`.
    """

    pass


@attrs.define
class Screenshot:
    """A container for holding the screenshot of a webpage."""

    #: Body.
    body: bytes

    @classmethod
    def from_base64(cls, body):
        return cls(body=b64decode(body.encode()))



import json
import re
from copy import deepcopy
from inspect import isclass
from typing import Any, Dict
from unittest import mock

import pytest
from pytest_twisted import ensureDeferred
from scrapy import Request, Spider
from scrapy.core.downloader.handlers.http import (
    HTTP10DownloadHandler,
    HTTPDownloadHandler,
)
from scrapy.exceptions import NotConfigured
from scrapy.settings import Settings
from scrapy.utils.misc import create_instance, load_object
from scrapy.utils.test import get_crawler
from zyte_api.aio.client import AsyncClient
from zyte_api.aio.retry import RetryFactory
from zyte_api.constants import API_URL

from scrapy_zyte_api.handler import (
    ScrapyZyteAPIDownloadHandler,
    ScrapyZyteAPIHTTPDownloadHandler,
)
from scrapy_zyte_api.utils import _ADDON_SUPPORT, USER_AGENT

from . import DEFAULT_CLIENT_CONCURRENCY, SETTINGS, UNSET
from . import get_crawler as get_crawler_zyte_api
from . import get_download_handler, make_handler, set_env
from .mockserver import MockServer


@pytest.mark.parametrize(
    "concurrency",
    (
        1,
        DEFAULT_CLIENT_CONCURRENCY,
        DEFAULT_CLIENT_CONCURRENCY + 1,
    ),
)
def test_concurrency_configuration(concurrency):
    settings = {
        **SETTINGS,
        "CONCURRENT_REQUESTS": concurrency,
    }
    crawler = get_crawler(settings_dict=settings)
    handler = ScrapyZyteAPIDownloadHandler(
        settings=crawler.settings,
        crawler=crawler,
    )
    assert handler._client.n_conn == concurrency
    assert handler._session.connector.limit == concurrency


@pytest.mark.parametrize(
    "env_var,setting,expected",
    (
        (
            UNSET,
            UNSET,
            NotConfigured,
        ),
        (
            "",
            UNSET,
            "",
        ),
        (
            "a",
            UNSET,
            "a",
        ),
        (
            UNSET,
            None,
            NotConfigured,
        ),
        (
            "",
            None,
            "",
        ),
        (
            "a",
            None,
            "a",
        ),
        (
            UNSET,
            "",
            NotConfigured,
        ),
        (
            "",
            "",
            "",
        ),
        (
            "a",
            "",
            "a",
        ),
        (
            UNSET,
            "b",
            "b",
        ),
        (
            "",
            "b",
            "b",
        ),
        (
            "a",
            "b",
            "b",
        ),
    ),
)
def test_api_key(env_var, setting, expected):
    env = {}
    if env_var is not UNSET:
        env["ZYTE_API_KEY"] = env_var
    settings = {}
    if setting is not UNSET:
        settings["ZYTE_API_KEY"] = setting
    with set_env(**env):
        crawler = get_crawler(settings_dict=settings)

        def build_hander():
            return create_instance(
                ScrapyZyteAPIDownloadHandler,
                settings=None,
                crawler=crawler,
            )

        if isclass(expected) and issubclass(expected, Exception):
            with pytest.raises(expected):
                handler = build_hander()
        else:
            handler = build_hander()
            assert handler._client.api_key == expected


@pytest.mark.parametrize(
    "setting,expected",
    (
        (
            UNSET,
            API_URL,
        ),
        (
            None,
            API_URL,
        ),
        (
            "",
            API_URL,
        ),
        (
            "a",
            "a",
        ),
        (
            "https://api.example.com",
            "https://api.example.com",
        ),
    ),
)
def test_api_url(setting, expected):
    settings = {"ZYTE_API_KEY": "a"}
    if setting is not UNSET:
        settings["ZYTE_API_URL"] = setting
    crawler = get_crawler(settings_dict=settings)
    handler = create_instance(
        ScrapyZyteAPIDownloadHandler,
        settings=None,
        crawler=crawler,
    )
    assert handler._client.api_url == expected


def test_custom_client():
    client = AsyncClient(api_key="a", api_url="b")
    crawler = get_crawler()
    handler = ScrapyZyteAPIDownloadHandler(crawler.settings, crawler, client)
    assert handler._client == client
    assert handler._client != AsyncClient(api_key="a", api_url="b")


RETRY_POLICY_A = RetryFactory().build()
RETRY_POLICY_B = RetryFactory().build()
assert RETRY_POLICY_A != RETRY_POLICY_B


@ensureDeferred
@pytest.mark.parametrize(
    "settings,meta,expected",
    [
        ({}, {}, None),
        (
            {"ZYTE_API_RETRY_POLICY": "tests.test_handler.RETRY_POLICY_A"},
            {},
            RETRY_POLICY_A,
        ),
        ({}, {"zyte_api_retry_policy": RETRY_POLICY_B}, RETRY_POLICY_B),
        (
            {},
            {"zyte_api_retry_policy": "tests.test_handler.RETRY_POLICY_B"},
            RETRY_POLICY_B,
        ),
        (
            {"ZYTE_API_RETRY_POLICY": "tests.test_handler.RETRY_POLICY_A"},
            {"zyte_api_retry_policy": RETRY_POLICY_B},
            RETRY_POLICY_B,
        ),
        (
            {"ZYTE_API_RETRY_POLICY": "tests.test_handler.RETRY_POLICY_A"},
            {"zyte_api_retry_policy": "tests.test_handler.RETRY_POLICY_B"},
            RETRY_POLICY_B,
        ),
    ],
)
async def test_retry_policy(
    settings: Dict[str, Any],
    meta: Dict[str, Any],
    expected: Any,
):
    meta = {"zyte_api": {"browserHtml": True}, **meta}
    async with make_handler(settings) as handler:
        req = Request("https://example.com", meta=meta)
        unmocked_client = handler._client
        handler._client = mock.AsyncMock(unmocked_client)
        handler._client.request_raw.return_value = {
            "browserHtml": "",
            "url": "",
        }
        await handler.download_request(req, None)

        # What we're interested in is the Request call in the API
        request_call = [
            c for c in handler._client.mock_calls if "request_raw(" in str(c)
        ]

        if not request_call:
            pytest.fail("The client's request_raw() method was not called.")

        actual = request_call[0].kwargs["retrying"]
        assert actual == expected


@ensureDeferred
async def test_stats(mockserver):
    async with make_handler({}, mockserver.urljoin("/")) as handler:
        scrapy_stats = handler._stats
        assert scrapy_stats.get_stats() == {}

        meta = {
            "zyte_api": {"a": "...", "b": {"b0": "..."}, "experimental": {"c0": "..."}}
        }
        request = Request("https://example.com", meta=meta)
        await handler.download_request(request, None)

        assert set(scrapy_stats.get_stats()) == {
            f"scrapy-zyte-api/{stat}"
            for stat in (
                "429",
                "attempts",
                "error_ratio",
                "errors",
                "fatal_errors",
                "mean_connection_seconds",
                "mean_response_seconds",
                "processed",
                "request_args/a",
                "request_args/b",
                "request_args/experimental.c0",
                "request_args/url",
                "status_codes/200",
                "success_ratio",
                "success",
                "throttle_ratio",
            )
        }
        for suffix, value in (
            ("429", 0),
            ("attempts", 1),
            ("error_ratio", 0.0),
            ("errors", 0),
            ("fatal_errors", 0),
            ("processed", 1),
            ("request_args/a", 1),
            ("request_args/b", 1),
            ("request_args/experimental.c0", 1),
            ("request_args/url", 1),
            ("status_codes/200", 1),
            ("success_ratio", 1.0),
            ("success", 1),
            ("throttle_ratio", 0.0),
        ):
            stat = f"scrapy-zyte-api/{suffix}"
            assert scrapy_stats.get_value(stat) == value
        for name in ("connection", "response"):
            stat = f"scrapy-zyte-api/mean_{name}_seconds"
            value = scrapy_stats.get_value(stat)
            assert isinstance(value, float)
            assert value > 0.0


def test_single_client():
    """Make sure that the same Zyte API client is used by both download
    handlers."""
    crawler = get_crawler(settings_dict=SETTINGS)
    handler1 = ScrapyZyteAPIDownloadHandler(
        settings=crawler.settings,
        crawler=crawler,
    )
    handler2 = ScrapyZyteAPIDownloadHandler(
        settings=crawler.settings,
        crawler=crawler,
    )
    assert handler1._client is handler2._client


@ensureDeferred
@pytest.mark.parametrize(
    "settings,enabled",
    [
        ({}, False),
        ({"ZYTE_API_LOG_REQUESTS": False}, False),
        ({"ZYTE_API_LOG_REQUESTS": True}, True),
    ],
)
async def test_log_request_toggle(
    settings: Dict[str, Any],
    enabled: bool,
    mockserver,
):
    async with make_handler(settings, mockserver.urljoin("/")) as handler:
        meta = {"zyte_api": {"foo": "bar"}}
        request = Request("https://example.com", meta=meta)
        with mock.patch("scrapy_zyte_api.handler.logger") as logger:
            await handler.download_request(request, None)
        if enabled:
            logger.debug.assert_called()
        else:
            logger.debug.assert_not_called()


@ensureDeferred
@pytest.mark.parametrize(
    "settings,short_str,long_str,truncated_str",
    [
        ({}, "a" * 64, "a" * 65, "a" * 63 + "..."),
        ({"ZYTE_API_LOG_REQUESTS_TRUNCATE": 0}, "a" * 64, "a" * 65, "a" * 65),
        ({"ZYTE_API_LOG_REQUESTS_TRUNCATE": 1}, "a", "aa", "..."),
        ({"ZYTE_API_LOG_REQUESTS_TRUNCATE": 2}, "aa", "aaa", "a..."),
    ],
)
async def test_log_request_truncate(
    settings: Dict[str, Any],
    short_str: str,
    long_str: str,
    truncated_str: str,
    mockserver,
):
    settings["ZYTE_API_LOG_REQUESTS"] = True
    input_params = {
        "short": short_str,
        "long": long_str,
        "list": [
            short_str,
            long_str,
            {
                "short": short_str,
                "long": long_str,
            },
        ],
        "dict": {
            "short": short_str,
            "long": long_str,
            "list": [
                short_str,
                long_str,
            ],
        },
    }
    expected_logged_params = {
        "short": short_str,
        "long": truncated_str,
        "list": [
            short_str,
            truncated_str,
            {
                "short": short_str,
                "long": truncated_str,
            },
        ],
        "dict": {
            "short": short_str,
            "long": truncated_str,
            "list": [
                short_str,
                truncated_str,
            ],
        },
    }
    expected_api_params = deepcopy(input_params)
    async with make_handler(settings, mockserver.urljoin("/")) as handler:
        meta = {"zyte_api": input_params}
        request = Request("https://example.com", meta=meta)
        unmocked_client = handler._client
        handler._client = mock.AsyncMock(unmocked_client)
        handler._client.request_raw.return_value = {
            "browserHtml": "",
            "url": "",
        }
        with mock.patch("scrapy_zyte_api.handler.logger") as logger:
            await handler.download_request(request, None)

        # Check that the logged params are truncated.
        logged_message = logger.debug.call_args[0][0]
        logged_params_json_match = re.search(r"\{.*", logged_message)
        assert logged_params_json_match is not None
        logged_params_json = logged_params_json_match[0]
        logged_params = json.loads(logged_params_json)
        del logged_params["url"]
        assert logged_params == expected_logged_params

        # Check that the actual params are *not* truncated.
        actual_api_params = handler._client.request_raw.call_args[0][0]
        del actual_api_params["url"]
        assert actual_api_params == expected_api_params


@pytest.mark.parametrize("enabled", [True, False])
def test_log_request_truncate_negative(enabled):
    settings: Dict[str, Any] = {
        **SETTINGS,
        "ZYTE_API_LOG_REQUESTS": enabled,
        "ZYTE_API_LOG_REQUESTS_TRUNCATE": -1,
    }
    crawler = get_crawler(settings_dict=settings)
    with pytest.raises(ValueError):
        create_instance(
            ScrapyZyteAPIDownloadHandler,
            settings=None,
            crawler=crawler,
        )


@pytest.mark.parametrize("enabled", [True, False, None])
def test_trust_env(enabled):
    settings: Dict[str, Any] = {
        **SETTINGS,
    }
    if enabled is not None:
        settings["ZYTE_API_USE_ENV_PROXY"] = enabled
    else:
        enabled = False
    crawler = get_crawler(settings_dict=settings)
    handler = create_instance(
        ScrapyZyteAPIDownloadHandler,
        settings=None,
        crawler=crawler,
    )
    assert handler._session._trust_env == enabled


@pytest.mark.parametrize(
    "user_agent,expected",
    (
        (
            None,
            USER_AGENT,
        ),
        (
            "zyte-crawlers/0.0.1",
            "zyte-crawlers/0.0.1",
        ),
    ),
)
def test_user_agent_for_build_client(user_agent, expected):
    settings = Settings(
        {
            **SETTINGS,
            "_ZYTE_API_USER_AGENT": user_agent,
        }
    )
    client = ScrapyZyteAPIDownloadHandler._build_client(settings)
    assert client.user_agent == expected


@ensureDeferred
async def test_bad_key():
    class TestSpider(Spider):
        name = "test"
        start_urls = ["https://bad-key.example"]

        def parse(self, response):
            pass

    settings = {
        "ZYTE_API_TRANSPARENT_MODE": True,
        **SETTINGS,
    }

    with MockServer() as server:
        settings["ZYTE_API_URL"] = server.urljoin("/")
        crawler = get_crawler(TestSpider, settings_dict=settings)
        await crawler.crawl()

    assert crawler.stats.get_value("finish_reason") == "zyte_api_bad_key"


# NOTE: Under the assumption that a case of bad key will happen since the
# beginning of a crawl, we only test the start_urls scenario, and not also the
# case of follow-up responses suddenly giving such an error.


@ensureDeferred
async def test_suspended_account_start_urls():
    class TestSpider(Spider):
        name = "test"
        start_urls = ["https://suspended-account.example"]

        def parse(self, response):
            pass

    settings = {
        "ZYTE_API_TRANSPARENT_MODE": True,
        **SETTINGS,
    }

    with MockServer() as server:
        settings["ZYTE_API_URL"] = server.urljoin("/")
        crawler = get_crawler(TestSpider, settings_dict=settings)
        await crawler.crawl()

    assert crawler.stats.get_value("finish_reason") == "zyte_api_suspended_account"


@ensureDeferred
async def test_suspended_account_callback():
    class TestSpider(Spider):
        name = "test"
        start_urls = ["https://example.com"]

        def parse(self, response):
            yield response.follow("https://suspended-account.example")

    settings = {
        "ZYTE_API_TRANSPARENT_MODE": True,
        **SETTINGS,
    }

    with MockServer() as server:
        settings["ZYTE_API_URL"] = server.urljoin("/")
        crawler = get_crawler(TestSpider, settings_dict=settings)
        await crawler.crawl()

    assert crawler.stats.get_value("finish_reason") == "zyte_api_suspended_account"


def test_fallback_setting():
    crawler = get_crawler_zyte_api(settings=SETTINGS)
    handler = get_download_handler(crawler, "https")
    assert isinstance(handler, ScrapyZyteAPIDownloadHandler)
    assert isinstance(handler._fallback_handler, HTTPDownloadHandler)


@pytest.mark.skipif(not _ADDON_SUPPORT, reason="No addon support in Scrapy < 2.10")
@ensureDeferred
async def test_addon(mockserver):
    async with make_handler({}, mockserver.urljoin("/"), use_addon=True) as handler:
        request = Request("https://example.com")
        await handler.download_request(request, None)
        assert handler._stats.get_value("scrapy-zyte-api/success") == 1


@pytest.mark.skipif(not _ADDON_SUPPORT, reason="No addon support in Scrapy < 2.10")
@ensureDeferred
async def test_addon_disable_transparent(mockserver):
    async with make_handler(
        {"ZYTE_API_TRANSPARENT_MODE": False}, mockserver.urljoin("/"), use_addon=True
    ) as handler:
        request = Request("https://toscrape.com")
        await handler.download_request(request, None)
        assert handler._stats.get_value("scrapy-zyte-api/success") is None

        meta = {"zyte_api": {"foo": "bar"}}
        request = Request("https://toscrape.com", meta=meta)
        await handler.download_request(request, None)
        assert handler._stats.get_value("scrapy-zyte-api/success") == 1


@pytest.mark.skipif(not _ADDON_SUPPORT, reason="No addon support in Scrapy < 2.10")
@ensureDeferred
async def test_addon_fallback():
    settings = {
        "DOWNLOAD_HANDLERS": {
            "http": "scrapy.core.downloader.handlers.http.HTTP10DownloadHandler"
        },
    }
    crawler = get_crawler_zyte_api(settings, use_addon=True)
    handler = get_download_handler(crawler, "http")
    assert isinstance(handler, ScrapyZyteAPIHTTPDownloadHandler)
    assert isinstance(handler._fallback_handler, HTTP10DownloadHandler)


@pytest.mark.skipif(not _ADDON_SUPPORT, reason="No addon support in Scrapy < 2.10")
@ensureDeferred
async def test_addon_fallback_explicit():
    settings = {
        "ZYTE_API_FALLBACK_HTTP_HANDLER": "scrapy.core.downloader.handlers.http.HTTP10DownloadHandler",
    }
    crawler = get_crawler_zyte_api(settings, use_addon=True)
    handler = get_download_handler(crawler, "http")
    assert isinstance(handler, ScrapyZyteAPIHTTPDownloadHandler)
    assert isinstance(handler._fallback_handler, HTTP10DownloadHandler)


@pytest.mark.skipif(not _ADDON_SUPPORT, reason="No addon support in Scrapy < 2.10")
@ensureDeferred
async def test_addon_matching_settings():
    def serialize(settings):
        result = dict(settings)
        for setting in (
            "ADDONS",
            "ZYTE_API_FALLBACK_HTTP_HANDLER",
            "ZYTE_API_FALLBACK_HTTPS_HANDLER",
        ):
            if setting in settings:
                del result[setting]
        for setting in (
            "DOWNLOADER_MIDDLEWARES",
            "SPIDER_MIDDLEWARES",
        ):
            for key in list(result[setting]):
                if isinstance(key, str):
                    obj = load_object(key)
                    result[setting][obj] = result[setting].pop(key)
        for key in result["DOWNLOAD_HANDLERS"]:
            result["DOWNLOAD_HANDLERS"][key] = result["DOWNLOAD_HANDLERS"][
                key
            ].__class__
        return result

    crawler = get_crawler_zyte_api({"ZYTE_API_TRANSPARENT_MODE": True})
    addon_crawler = get_crawler_zyte_api(use_addon=True)
    assert serialize(crawler.settings) == serialize(addon_crawler.settings)


@pytest.mark.skipif(not _ADDON_SUPPORT, reason="No addon support in Scrapy < 2.10")
@ensureDeferred
async def test_addon_custom_fingerprint():
    class CustomRequestFingerprinter:
        pass

    crawler = get_crawler_zyte_api(
        {"REQUEST_FINGERPRINTER_CLASS": CustomRequestFingerprinter}, use_addon=True
    )
    assert (
        crawler.settings["ZYTE_API_FALLBACK_REQUEST_FINGERPRINTER_CLASS"]
        == CustomRequestFingerprinter
    )
    assert (
        crawler.settings["REQUEST_FINGERPRINTER_CLASS"]
        == "scrapy_zyte_api.ScrapyZyteAPIRequestFingerprinter"
    )



from base64 import b64encode
from collections import defaultdict
from functools import partial
from typing import Any, Dict, cast

import pytest
from scrapy import Request
from scrapy.exceptions import NotSupported
from scrapy.http import Response, TextResponse
from scrapy.http.cookies import CookieJar

from scrapy_zyte_api.responses import (
    _API_RESPONSE,
    ZyteAPIResponse,
    ZyteAPITextResponse,
)
from scrapy_zyte_api.responses import _process_response as _unwrapped_process_response
from scrapy_zyte_api.utils import _RESPONSE_HAS_IP_ADDRESS, _RESPONSE_HAS_PROTOCOL

PAGE_CONTENT = "<html><body>The cake is a lie!</body></html>"
PAGE_CONTENT_2 = "<html><body>Ceci n’est pas une pipe</body></html>"
URL = "https://example.com"


INPUT_COOKIES = [
    {
        "name": "a",
        "value": "b",
        "domain": ".example.com",
        "path": "/",
        "expires": 1679893056,
        "httpOnly": True,
        "secure": True,
    },
    {
        "name": "c",
        "value": "d",
    },
]
OUTPUT_COOKIE_HEADERS = {
    b"Set-Cookie": [
        (
            b"a=b; "
            b"Domain=.example.com; "
            b"Path=/; "
            b"Expires=Mon, 27 Mar 2023 04:57:36 GMT; "
            b"HttpOnly; "
            b"Secure"
        ),
        (b"c=d"),
    ]
}

_process_response = partial(
    _unwrapped_process_response, cookie_jars=defaultdict(CookieJar)
)


def raw_api_response_browser():
    return {
        "url": URL,
        "browserHtml": PAGE_CONTENT,
        "javascript": True,
        "echoData": {"some_value": "here"},
        "httpResponseHeaders": [
            {"name": "Content-Type", "value": "text/html"},
            {"name": "Content-Length", "value": len(PAGE_CONTENT)},
        ],
        "statusCode": 200,
        "experimental": {
            "responseCookies": INPUT_COOKIES,
        },
    }


def raw_api_response_body():
    return {
        "url": "https://example.com",
        "httpResponseBody": b64encode(PAGE_CONTENT.encode("utf-8")),
        "echoData": {"some_value": "here"},
        "httpResponseHeaders": [
            {"name": "Content-Type", "value": "text/html"},
            {"name": "Content-Length", "value": len(PAGE_CONTENT)},
        ],
        "statusCode": 200,
        "experimental": {
            "responseCookies": INPUT_COOKIES,
        },
    }


def raw_api_response_mixed():
    return {
        "url": URL,
        "browserHtml": PAGE_CONTENT,
        "httpResponseBody": b64encode(PAGE_CONTENT_2.encode("utf-8")),
        "echoData": {"some_value": "here"},
        "httpResponseHeaders": [
            {"name": "Content-Type", "value": "text/html"},
            {"name": "Content-Length", "value": len(PAGE_CONTENT_2)},
        ],
        "statusCode": 200,
        "experimental": {
            "responseCookies": INPUT_COOKIES,
        },
    }


EXPECTED_BODY = PAGE_CONTENT.encode("utf-8")


@pytest.mark.parametrize(
    "api_response,cls",
    [
        (raw_api_response_browser, ZyteAPITextResponse),
        (raw_api_response_body, ZyteAPIResponse),
    ],
)
def test_init(api_response, cls):
    response = cls(URL, raw_api_response=api_response())
    assert response.raw_api_response == api_response()

    assert response.url == URL
    assert response.status == 200
    assert not response.headers
    assert response.body == b""
    assert not response.flags
    assert response.request is None
    assert response.certificate is None
    if _RESPONSE_HAS_IP_ADDRESS:
        assert response.ip_address is None
    if _RESPONSE_HAS_PROTOCOL:
        assert response.protocol is None


@pytest.mark.parametrize(
    "api_response,cls,content_length",
    [
        (raw_api_response_browser, ZyteAPITextResponse, 44),
        (raw_api_response_body, ZyteAPIResponse, 44),
        (raw_api_response_mixed, ZyteAPITextResponse, 49),
    ],
)
def test_text_from_api_response(api_response, cls, content_length):
    response = cls.from_api_response(api_response())
    assert response.raw_api_response == api_response()

    assert response.url == URL
    assert response.status == 200
    expected_headers = {
        b"Content-Type": [b"text/html"],
        b"Content-Length": [str(content_length).encode()],
        **OUTPUT_COOKIE_HEADERS,
    }
    assert response.headers == expected_headers
    assert response.body == EXPECTED_BODY
    assert response.flags == ["zyte-api"]
    assert response.request is None
    assert response.certificate is None
    if _RESPONSE_HAS_IP_ADDRESS:
        assert response.ip_address is None
    if _RESPONSE_HAS_PROTOCOL:
        assert response.protocol is None


@pytest.mark.parametrize(
    "api_response,cls",
    [
        (raw_api_response_browser, ZyteAPITextResponse),
        (raw_api_response_body, ZyteAPIResponse),
    ],
)
def test_response_replace(api_response, cls):
    orig_response = cls.from_api_response(api_response())

    # It should still work the same way
    new_response = orig_response.replace(status=404)
    assert new_response.status == 404

    new_response = orig_response.replace(url="https://new-example.com")
    assert new_response.url == "https://new-example.com"

    # Ensure that the Zyte API response is intact
    assert new_response.raw_api_response == api_response()

    new_raw_api_response = {
        "url": "https://another-website.com",
        "httpResponseHeaders": {"name": "Content-Type", "value": "application/json"},
    }

    # Attempting to replace the raw_api_response value would raise an error
    with pytest.raises(ValueError):
        orig_response.replace(raw_api_response=new_raw_api_response)


def test_non_utf8_response():
    content = "<html><body>Some non-ASCII ✨ chars</body></html>"
    sample_raw_api_response = {
        "url": URL,
        "browserHtml": content,
        "httpResponseHeaders": [
            {"name": "Content-Type", "value": "text/html; charset=iso-8859-1"},
            {"name": "Content-Length", "value": len(content)},
        ],
    }

    # Encoding inference should not kick in under the hood for
    # ``scrapy.http.TextResponse`` since ``ZyteAPITextResponse`` using "utf-8"
    # for it. This is the default encoding for the "browserHtml" contents from
    # Zyte API. Thus, even if the Response Headers or <meta> tags indicate a
    # different encoding, it should still be treated as "utf-8".
    response = ZyteAPITextResponse.from_api_response(sample_raw_api_response)
    assert response.text == content
    assert response.encoding == "utf-8"


BODY = "<html><body>Hello<h1>World!✨</h1></body></html>"


def format_to_httpResponseBody(body, encoding="utf-8"):
    return b64encode(body.encode(encoding)).decode("utf-8")


@pytest.mark.parametrize(
    "api_response,cls",
    [
        (raw_api_response_browser, ZyteAPITextResponse),
        (raw_api_response_body, ZyteAPIResponse),
    ],
)
def test_response_headers_removal(api_response, cls):
    """Headers like 'Content-Encoding' should be removed later in the response
    instance returned to Scrapy.

    However, they should still be present inside 'raw_api_response.headers'.
    """
    additional_headers = [
        {"name": "Content-Encoding", "value": "gzip"},
        {"name": "Set-Cookie", "value": "a=b"},
        {"name": "X-Some-Other-Value", "value": "123"},
    ]
    raw_response = api_response()
    raw_response["httpResponseHeaders"] = additional_headers

    response = cls.from_api_response(raw_response)

    expected_headers = {
        b"X-Some-Other-Value": [b"123"],
        **OUTPUT_COOKIE_HEADERS,
    }
    assert response.headers == expected_headers
    assert (
        response.raw_api_response["httpResponseHeaders"]
        == raw_response["httpResponseHeaders"]
    )


INPUT_COOKIES_SIMPLE = [{"name": "c", "value": "d"}]


@pytest.mark.parametrize(
    "fields,cls,keep",
    [
        # Only keep the Set-Cookie header if experimental.responseCookies is
        # not received.
        *(
            (
                {
                    **cast(Dict[Any, Any], output_fields),
                    "httpResponseHeaders": [
                        {"name": "Content-Type", "value": "text/html"},
                        {"name": "Content-Length", "value": str(len(PAGE_CONTENT))},
                    ],
                    **cookie_fields,  # type: ignore[dict-item]
                },
                response_cls,
                keep,
            )
            for output_fields, response_cls in (
                (
                    {"httpResponseBody": b64encode(PAGE_CONTENT.encode("utf-8"))},
                    ZyteAPIResponse,
                ),
                (
                    {
                        "browserHtml": PAGE_CONTENT,
                    },
                    ZyteAPITextResponse,
                ),
            )
            for cookie_fields, keep in (
                # No response cookies, so Set-Cookie is kept.
                (
                    {},
                    True,
                ),
                # Response cookies, so Set-Cookie is not kept.
                (
                    {
                        "experimental": {
                            "responseCookies": INPUT_COOKIES_SIMPLE,
                        },
                    },
                    False,
                ),
            )
        ),
    ],
)
def test_response_cookie_header(fields, cls, keep):
    """Test the logic to keep or not the Set-Cookie header in response
    headers."""
    expected_headers = {
        **{
            header["name"].encode(): [header["value"].encode()]
            for header in fields["httpResponseHeaders"]
        },
    }
    if keep:
        expected_headers[b"Set-Cookie"] = [b"a=b"]
    elif "experimental" in fields:
        expected_headers[b"Set-Cookie"] = [b"c=d"]

    fields["url"] = "https://example.com"
    fields["statusCode"] = 200
    fields["httpResponseHeaders"].append({"name": "Set-Cookie", "value": "a=b"})

    response = cls.from_api_response(fields)

    assert response.headers == expected_headers
    assert (
        response.raw_api_response["httpResponseHeaders"]
        == fields["httpResponseHeaders"]
    )


def test__process_response_no_body():
    """The _process_response() function should handle missing 'browserHtml' or
    'httpResponseBody'.
    """
    api_response: _API_RESPONSE = {
        "url": "https://example.com",
        "product": {"name": "shoes"},
    }

    resp = _process_response(api_response, Request(api_response["url"]))

    assert isinstance(resp, Response)
    assert resp.body == b""


def test__process_response_body_only():
    """Having the Body but with no Headers won't allow us to decode the contents
    with the proper encoding.

    Thus, we won't have access to css/xpath selectors.
    """
    encoding = "utf-8"
    api_response = {
        "url": "https://example.com",
        "httpResponseBody": format_to_httpResponseBody(BODY, encoding=encoding),
    }

    resp = _process_response(api_response, Request(api_response["url"]))

    assert isinstance(resp, Response)
    with pytest.raises(NotSupported):
        assert resp.css("h1 ::text")
    with pytest.raises(NotSupported):
        assert resp.xpath("//body/text()")


@pytest.mark.xfail(reason="encoding inference is not supported for now")
def test__process_response_body_only_infer_encoding():
    """The ``scrapy.TextResponse`` class has the ability to check the encoding
    by inferring it in the HTML body.

    However, this is a bit tricky since we need to somehow ensure that the body
    we're receiving is "text/html". We can't fully determine that without the
    headers.
    """
    encoding = "gb18030"
    body = (
        "<html>"
        '<head><meta http-equiv="Content-Type" content="text/html; charset="gb2312"></head>'
        "<body>Some ✨ contents</body>"
        "</html>"
    )

    api_response = {
        "url": "https://example.com",
        "httpResponseBody": format_to_httpResponseBody(body, encoding=encoding),
    }

    resp = _process_response(api_response, Request(api_response["url"]))

    assert isinstance(resp, TextResponse)
    assert resp.css("body ::text").get() == "Some ✨ contents"
    assert resp.xpath("//body/text()").getall() == ["Some ✨ contents"]


@pytest.mark.parametrize(
    "encoding,content_type",
    [
        ("utf-8", "text/html; charset=UTF-8"),
        ("gb18030", "text/html; charset=gb2312"),
    ],
)
def test__process_response_body_and_headers(encoding, content_type):
    """Having access to the Headers allow us to properly decode the contents
    and will have access to the css/xpath selectors.
    """
    api_response = {
        "url": "https://example.com",
        "httpResponseBody": format_to_httpResponseBody(BODY, encoding=encoding),
        "httpResponseHeaders": [{"name": "Content-Type", "value": content_type}],
    }

    resp = _process_response(api_response, Request(api_response["url"]))

    assert isinstance(resp, TextResponse)
    assert resp.css("h1 ::text").get() == "World!✨"
    assert resp.xpath("//body/text()").getall() == ["Hello"]
    assert resp.encoding == encoding


@pytest.mark.parametrize(
    "body,expected,actual_encoding,inferred_encoding",
    [
        ("<html><body>plain</body></html>", "plain", "cp1252", "cp1252"),
        (
            "<html><body>✨</body></html>",
            "✨",
            "utf-8",
            "utf-8",
        ),
        (
            "<html><body>✨</body></html>",
            "✨",
            "utf-16",
            "utf-16-le",
        ),
        (
            """<html><head><meta http-equiv="Content-Type" content="text/html; charset="gb2312">
            </head><body>✨</body></html>""",
            "✨",
            "gb18030",
            None,
        ),
    ],
)
def test__process_response_body_and_headers_but_no_encoding(
    body, expected, actual_encoding, inferred_encoding
):
    """Should both the body and headers are present but no 'Content-Type' encoding
    can be derived, it should infer from the body contents.
    """
    api_response = {
        "url": "https://example.com",
        "httpResponseBody": format_to_httpResponseBody(body, encoding=actual_encoding),
        "httpResponseHeaders": [{"name": "X-Value", "value": "some_value"}],
    }

    resp = _process_response(api_response, Request(api_response["url"]))

    assert isinstance(resp, TextResponse)

    if inferred_encoding:
        assert resp.css("body ::text").get() == expected
        assert resp.xpath("//body/text()").get() == expected
        assert resp.encoding == inferred_encoding

    # Scrapy's ``TextResponse`` built-in inference only works on "utf-8" and
    # "Latin-1" based encodings.
    else:
        assert resp.css("body ::text").get() != expected
        assert resp.xpath("//body/text()").get() != expected
        assert resp.encoding == "ascii"


def test__process_response_body_and_headers_mismatch():
    """If the actual contents have a mismatch in terms of its encoding, we won't
    properly decode the ✨ emoji.
    """
    encoding = "utf-8"
    api_response = {
        "url": "https://example.com",
        "httpResponseBody": format_to_httpResponseBody(BODY, encoding=encoding),
        "httpResponseHeaders": [
            {"name": "Content-Type", "value": "text/html; charset=gb2312"}
        ],
    }

    resp = _process_response(api_response, Request(api_response["url"]))

    assert isinstance(resp, TextResponse)
    assert resp.css("h1 ::text").get() != "World!✨"  # mismatch
    assert resp.xpath("//body/text()").getall() == ["Hello"]
    assert resp.encoding == "gb18030"


def test__process_response_non_text():
    """Non-textual responses like images, files, etc. won't have access to the
    css/xpath selectors.
    """
    api_response: _API_RESPONSE = {
        "url": "https://example.com/sprite.gif",
        "httpResponseBody": "",
        "httpResponseHeaders": [
            {
                "name": "Content-Type",
                "value": "image/gif",
            }
        ],
    }
    resp = _process_response(api_response, Request(api_response["url"]))

    assert isinstance(resp, Response)
    with pytest.raises(NotSupported):
        assert resp.css("h1 ::text")
    with pytest.raises(NotSupported):
        assert resp.xpath("//body/text()")


@pytest.mark.parametrize(
    "api_response",
    [
        {"url": "https://example.com", "browserHtml": BODY},
        {
            "url": "https://example.com",
            "browserHtml": BODY,
            "httpResponseHeaders": [
                {
                    "name": "Content-Type",
                    "value": "text/html; charset=UTF-8",
                }
            ],
        },
    ],
)
def test__process_response_browserhtml(api_response):
    resp = _process_response(api_response, Request(api_response["url"]))

    assert isinstance(resp, TextResponse)
    assert resp.css("h1 ::text").get() == "World!✨"
    assert resp.xpath("//body/text()").getall() == ["Hello"]
    assert resp.encoding == "utf-8"  # Zyte API is consistent with this on browserHtml


@pytest.mark.parametrize(
    "base_kwargs_func",
    [
        raw_api_response_browser,
        raw_api_response_body,
    ],
)
@pytest.mark.parametrize(
    "kwargs,expected_status_code",
    [
        ({}, 200),
        ({"statusCode": 200}, 200),
        ({"statusCode": 404}, 404),
    ],
)
def test_status_code(base_kwargs_func, kwargs, expected_status_code):
    base_api_response = base_kwargs_func()
    del base_api_response["statusCode"]
    api_response = {**base_api_response, **kwargs}
    response = _process_response(api_response, Request(api_response["url"]))
    assert response is not None
    assert response.status == expected_status_code
